"""_fetchrepodata: Map yum/dnf repo metadata to local lookup caches"""
import gzip
import json
import logging
import os
from collections import defaultdict

import click
import requests

from gi.repository import Modulemd

from attr import attributes, attrib
from lxml import etree
from requests_toolbelt.downloadutils.tee import tee_to_file
from urllib.parse import urljoin

XDG_CACHE_HOME = os.environ.get("XDG_CACHE_HOME") or os.path.expanduser("~/.cache")
CACHEDIR = os.path.join(XDG_CACHE_HOME, "fedmod")

log = logging.getLogger(__name__)

FALLBACK_STREAM = 'master'
STREAM = 'f28'
ARCH = 'x86_64'
_F28_MODULAR_REPO = "https://dl.fedoraproject.org/pub/fedora/linux/releases/28/Modular/"
_F28_MAIN_REPO = "https://dl.fedoraproject.org/pub/fedora/linux/releases/28/Everything/"
_F28_UPDATES_REPO = "https://dl.fedoraproject.org/pub/fedora/linux/updates/28/Everything/"
#_F28_MODULAR_UPDATES_REPO = "https://dl.fedoraproject.org/pub/fedora/linux/updates/28/Modular/"

class MissingMetadata(Exception):
    """Reports failure to find the local metadata cache"""

@attributes
class RepoPaths:
    remote_repo_url = attrib(str)
    remote_metadata_url = attrib(str)
    local_cache_path = attrib(str)
    local_metadata_path = attrib(str)

def _define_repo(remote_prefix, local_cache_name, arch=None):
    if arch is None:
        local_arch_path = "source"
        if "updates" in remote_prefix:
            remote_arch_path = "SRPMS"
        else:
            remote_arch_path = "source/tree/"
    else:
        local_arch_path = arch
        if "updates" in remote_prefix:
            remote_arch_path = arch + "/"
        else:
            remote_arch_path = os.path.join(arch, "os/")
    remote_repo_url = os.path.join(remote_prefix, remote_arch_path)
    remote_metadata_url = os.path.join(remote_repo_url, "repodata/")
    local_cache_path = os.path.join(CACHEDIR, "repos", local_cache_name, local_arch_path)
    local_metadata_path = os.path.join(local_cache_path, "repodata/")
    return RepoPaths(remote_repo_url, remote_metadata_url,
                     local_cache_path, local_metadata_path)

_x86_64_MODULE_INFO = _define_repo(_F28_MODULAR_REPO, "f28-modular", ARCH)
_SOURCE_MODULE_INFO = _define_repo(_F28_MODULAR_REPO, "f28-modular")
#_x86_64_MODULE_UPDATES_INFO = _define_repo(_F28_MODULAR_UPDATES_REPO, "f28-modular-updates", ARCH)
#_SOURCE_MODULE_UPDATES_INFO = _define_repo(_F28_MODULAR_UPDATES_REPO, "f28-modular-updates")
_x86_64_PACKAGE_INFO = _define_repo(_F28_MAIN_REPO, "f28-packages", ARCH)
_SOURCE_PACKAGE_INFO = _define_repo(_F28_MAIN_REPO, "f28-packages")
_x86_64_UPDATES_INFO = _define_repo(_F28_UPDATES_REPO, "f28-updates", ARCH)
_SOURCE_UPDATES_INFO = _define_repo(_F28_UPDATES_REPO, "f28-updates")
_ALL_REPOS = (
    _x86_64_MODULE_INFO,
    _SOURCE_MODULE_INFO,
#    _x86_64_MODULE_UPDATES_INFO,
#    _SOURCE_MODULE_UPDATES_INFO,
    _x86_64_PACKAGE_INFO,
    _SOURCE_PACKAGE_INFO,
    _x86_64_UPDATES_INFO,
    _SOURCE_UPDATES_INFO,
)

_LOOKUP_CACHES = {
    "_MODULE_FORWARD_LOOKUP_CACHE": os.path.join(CACHEDIR, "f28-module-contents-cache.json"),
    "_SRPM_REVERSE_LOOKUP_CACHE": os.path.join(CACHEDIR, "f28-srpm-to-module-cache.json"),
    "_RPM_REVERSE_LOOKUP_CACHE": os.path.join(CACHEDIR, "f28-rpm-to-module-cache.json"),
}

METADATA_SECTIONS = ("filelists", "primary", "modules")

_REPOMD_XML_NAMESPACE = {"rpm": "http://linux.duke.edu/metadata/repo"}
def _read_repomd_location(repomd_xml, section):
    location = repomd_xml.find(f"rpm:data[@type='{section}']/rpm:location", _REPOMD_XML_NAMESPACE)
    if location is not None:
        return location.attrib["href"]
    return None

def _download_one_file(remote_url, filename):
    if os.path.exists(filename) and not filename.endswith((".xml", ".yaml")):
        print(f"  Skipping download; {filename} already exists")
        return
    response = requests.get(remote_url, stream=True)
    try:
        print(f"  Downloading {remote_url}")
        chunksize = 65536
        expected_chunks = int(response.headers["content-length"]) / chunksize
        downloader = tee_to_file(response, filename=filename, chunksize=chunksize)
        show_progress = click.progressbar(downloader, length=expected_chunks)
        with show_progress:
            for chunk in show_progress:
                pass
    finally:
        response.close()
    print(f"  Added {filename} to cache")

def _download_metadata_files(repo_paths):
    local_path = repo_paths.local_cache_path
    local_metadata_path = repo_paths.local_metadata_path
    os.makedirs(local_metadata_path, exist_ok=True)
    repomd_url = urljoin(repo_paths.remote_metadata_url, "repomd.xml")
    print(f"Remote metadata: {repomd_url}")
    response = requests.get(repomd_url)
    response.raise_for_status()
    repomd_filename = os.path.join(local_metadata_path, "repomd.xml")
    with open(repomd_filename, "wb") as f:
        f.write(response.content)
    print(f"  Cached metadata in {repomd_filename}")
    repomd_xml = etree.parse(repomd_filename)
    files_to_fetch = set()
    for section in METADATA_SECTIONS:
        relative_href = _read_repomd_location(repomd_xml, section)
        if relative_href is not None:
            files_to_fetch.add(relative_href)
    predownload = set(os.listdir(local_path))
    for relative_href in files_to_fetch:
        absolute_href = urljoin(repo_paths.remote_repo_url, relative_href)
        filename = os.path.join(local_path, relative_href)
        # This could be parallelised with concurrent.futures, but
        # probably not worth it (it makes the progress bars trickier)
        _download_one_file(absolute_href, filename)
    postdownload = set(os.listdir(local_path))
    # Prune any old metadata files automatically
    if len(postdownload) >= (len(predownload) + len(METADATA_SECTIONS)):
        # TODO: Actually prune old metadata files
        pass

def _write_cache(cache_name, data):
    """Write the given data to the nominated cache file"""
    cache_fname = _LOOKUP_CACHES[cache_name]
    with open(cache_fname, "w") as cache_file:
        json.dump(data, cache_file)
    print(f"  Added {cache_fname} to cache")

def _read_cache(cache_name):
    """Read the parsed data from the nominated cache file"""
    cache_fname = _LOOKUP_CACHES[cache_name]
    with open(cache_fname, "r") as cache_file:
        return json.load(cache_file)

def _write_lookup_caches():
    metadata_dir = os.path.join(_x86_64_MODULE_INFO.local_cache_path)
    repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")
    repomd_xml = etree.parse(repomd_fname)
    repo_relative_modulemd = _read_repomd_location(repomd_xml, "modules")
    repo_modulemd_fname = os.path.join(metadata_dir, repo_relative_modulemd)
    with gzip.open(repo_modulemd_fname, "rt") as modules_yaml_gz:
        modules_yaml = modules_yaml_gz.read()
    module_forward_lookup = {}
    srpm_reverse_lookup = defaultdict(list)
    rpm_reverse_lookup = defaultdict(list)

    objects = Modulemd.objects_from_string(modules_yaml)
    for module in (o for o in objects if isinstance(o,  Modulemd.Module)):
        module_name = module.props.name
        artifacts = module.props.rpm_artifacts.get()

        module_forward_lookup[module_name] = list(set(artifacts))
        components = module.props.components_rpm
        for srpmname in components:
            srpm_reverse_lookup[srpmname].append(module_name)
        for rpmname in artifacts:
            rpmprefix = rpmname.split(":", 1)[0].rsplit("-", 1)[0]
            rpm_reverse_lookup[rpmprefix].append(module_name)
    # Cache the lookup tables as local JSON files
    print("Caching lookup tables")
    _write_cache("_MODULE_FORWARD_LOOKUP_CACHE", module_forward_lookup)
    _write_cache("_SRPM_REVERSE_LOOKUP_CACHE", srpm_reverse_lookup)
    _write_cache("_RPM_REVERSE_LOOKUP_CACHE", rpm_reverse_lookup)


def download_repo_metadata():
    """Downloads the latest repo metadata"""
    for repo_definition in _ALL_REPOS:
        _download_metadata_files(repo_definition)
    _write_lookup_caches()

@attributes
class LocalMetadataCache:
    dataset_name = attrib(str)
    cache_dir = attrib(str)
    srpm_to_modules = attrib(dict)
    rpm_to_modules = attrib(dict)
    module_to_packages = attrib(dict)
    source_repo_cache = attrib(str)
    arch_repo_cache = attrib(str)
    source_updates_cache = attrib(str)
    arch_updates_cache = attrib(str)


def load_cached_repodata(dataset_name):
    if dataset_name != "f28-x86_64":
        raise RuntimeError("Data sets other than 'f28-x86_64' are not yet supported")
    # Check whether or not fetch-metadata has been run at all
    metadata_dir = os.path.join(_x86_64_MODULE_INFO.local_cache_path)
    repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")
    if not os.path.exists(repomd_fname):
        msg = f"{repomd_fname!r} does not exist. Run `fedmod fetch-metadata`."
        raise MissingMetadata(msg)
    # Check whether or not fetch-metadata actually finished
    for cache_entry in _LOOKUP_CACHES.values():
        if not os.path.exists(cache_entry):
            msg = (f"{cache_entry!r} does not exist. "
                    "Try running `fedmod fetch-metadata` again.")
            raise MissingMetadata(msg)
    # Load the metadata
    return LocalMetadataCache(
        dataset_name = dataset_name,
        cache_dir = CACHEDIR,
        srpm_to_modules = _read_cache("_SRPM_REVERSE_LOOKUP_CACHE"),
        rpm_to_modules = _read_cache("_RPM_REVERSE_LOOKUP_CACHE"),
        module_to_packages = _read_cache("_MODULE_FORWARD_LOOKUP_CACHE"),
        source_repo_cache = _SOURCE_PACKAGE_INFO.local_cache_path,
        arch_repo_cache = _x86_64_PACKAGE_INFO.local_cache_path,
        source_updates_cache = _SOURCE_UPDATES_INFO.local_cache_path,
        arch_updates_cache = _x86_64_UPDATES_INFO.local_cache_path,
    )
